# -*- coding: UTF-8 -*-

from collections import deque
from typing import Callable, Optional, Any, Deque
from xml import sax

from wpydumps.model import Page, Revision, Contributor
from wpydumps.archive import get_reader

PageCallbackType = Callable[[Page], Any]


class PageHandler(sax.handler.ContentHandler):
    """
    SAX handler to parse page revisions from a Wikipedia dump.
    It calls ``page_callback`` on each ``wpydumps.model.Page`` it parses.

    Options:

    * ``keep_revisions_text`` (default ``False``): keep the revision text. If
      ``False``, drop it and set only `Revision.text_length`.

    Format: https://en.wikipedia.org/wiki/Help:Export.
    """

    def __init__(self, page_callback: PageCallbackType, keep_revisions_text=False):
        super().__init__()
        self.page_callback = page_callback
        self._keep_revisions_text = keep_revisions_text

        self._elements: Deque[str] = deque()
        self._current_page: Optional[Page] = None
        self._current_revision: Optional[Revision] = None
        self._current_contributor: Optional[Contributor] = None

        self._previous_revision_text_length: Optional[int] = None

    def _printState(self):
        print(list(self._elements))

    def currentElement(self):
        if self._elements:
            return self._elements[-1]

    def parentElement(self):
        if len(self._elements) >= 2:
            return self._elements[-2]

    def startElement(self, name, attrs):
        self._elements.append(name)

        if name == "page":
            self._current_page = Page()
            self._previous_revision_text_length = 0
            return

        parent = self.parentElement()
        if parent == "page":
            if name == "redirect":
                self._current_page.redirect = attrs["title"]
            elif name == "revision":
                self._current_revision = Revision()

        elif parent == "revision":
            if name == "minor":
                self._current_revision.minor = True
            elif name == "contributor":
                if "deleted" in attrs:
                    self._current_revision.deleted_contributor = True
                else:
                    self._current_contributor = Contributor()
            elif name == "text" and "deleted" in attrs:
                self._current_revision.deleted_text = True

    def characters(self, content):
        parent = self.parentElement()
        element = self.currentElement()
        revision = self._current_revision

        if content.isspace():
            if parent == "revision" and element == "text" and self._keep_revisions_text:
                if revision.text is None:
                    revision.text = ""
                revision.text += content
            return

        if parent == "page":
            page = self._current_page

            if element == "ns":
                page.namespace = content
            elif element == "id":
                page.id = content
            elif element == "restrictions":
                page.restrictions = content
            elif element == "title":
                page.title = content

        elif parent == "revision":

            if element == "timestamp":
                revision.timestamp = content
            elif element == "comment":
                if revision.comment is None:
                    revision.comment = ""
                revision.comment += content
            elif element == "text":
                if self._keep_revisions_text:
                    if revision.text is None:
                        revision.text = ""
                    revision.text += content

                if revision.text_length is None:
                    revision.text_length = 0

                revision.text_length += len(content)

            elif element == "id":
                revision.id = content
            elif element == "parentid":
                revision.parent_id = content
            elif element == "model":
                revision.model = content
            elif element == "format":
                revision.format = content
            elif element == "sha1":
                revision.sha1 = content

        elif parent == "contributor":
            contributor = self._current_contributor

            if element == "id":
                contributor.id = content
            elif element == "username":
                contributor.username = content
            elif element == "ip":
                contributor.ip = content

    def endElement(self, name):
        self._elements.pop()

        if name == "page":
            self.page_callback(self._current_page)
            self._current_page = None

        elif name == "revision":
            revision = self._current_revision
            if revision.text_length is None:
                revision.text_length = 0

            revision.diff_length = revision.text_length - self._previous_revision_text_length
            self._previous_revision_text_length = revision.text_length

            self._current_page.revisions.append(revision)
            self._current_revision = None

        elif name == "contributor":
            self._current_revision.contributor = self._current_contributor
            self._current_contributor = None

        # elif self.currentElement() == "revision":
        #     revision = self._current_revision


def parse_pages_from_reader(reader, page_callback: PageCallbackType,
                            keep_revisions_text=False):
    """
    Parse pages from a file-like reader. Call ``page_callback`` on each parsed
    page. See ``PageHandler`` for the keyword arguments.
    """
    handler = PageHandler(page_callback, keep_revisions_text=keep_revisions_text)
    sax.parse(reader, handler)


def parse_pages_from_archive_filename(filename: str, page_callback: PageCallbackType,
                                      keep_revisions_text=False):
    """
    Parse pages from an archive file. See ``parse_pages_from_reader``.
    """
    return parse_pages_from_reader(get_reader(filename), page_callback,
                                   keep_revisions_text=keep_revisions_text)



# -*- coding: UTF-8 -*-

import io
from typing import Iterable

import libarchive.public


class UnSupportedFileTypeException(Exception):
    def __init__(self, filename: str, extensions: Iterable[str]):
        super().__init__(f"Unsupported file type: {filename}. Supported extensions: {', '.join(extensions)}.")


class Wp7zReader(io.RawIOBase):
    """
    File-like reader that can read 7z archives. It extract each entry of the
    archive one-by-one, with no separation between them. Wikipedia dumps always
    contain only one entry.

    This is a wrapper around libarchive.public.file_reader.
    """

    def __init__(self, filename: str):
        self.filename = filename
        self._blocks = self._open_blocks()
        self._block = None
        self._cursor = 0

    def readable(self):
        return True

    def writable(self):
        return False

    def seekable(self):
        return False

    def close(self):
        pass

    def readinto(self, b):
        if not self._block or self._cursor == len(self._block):
            self._next_block()
            if not self._block:
                # EOF
                return 0

        size = min(len(b), len(self._block) - self._cursor)
        b[0:size] = self._block[self._cursor:self._cursor + size]
        self._cursor += size
        return size

    def _next_block(self):
        try:
            self._block = next(self._blocks)
            self._cursor = 0
        except StopIteration:
            return

    def _open_blocks(self):
        with libarchive.public.file_reader(self.filename) as entries:
            for entry in entries:
                for block in entry.get_blocks():
                    yield block
                break

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return


def get_reader(filename: str):
    """
    Get a reader based on the file extension.
    """
    if filename.endswith(".7z"):
        return Wp7zReader(filename)

    # Note those are here for convenience, but using 'with open' by yourself works as well.
    if filename.endswith(".bz2"):
        import bz2
        return bz2.open(filename, "rb")

    if filename.endswith(".xml"):
        return open(filename)

    raise UnSupportedFileTypeException(filename, (".7z", ".bz2"))



# -*- coding: UTF-8 -*-


def _repr(obj, attrs_string):
    name = "%s.%s" % (obj.__class__.__module__, obj.__class__.__name__)
    return '<%s %s>' % (name, attrs_string)


class Contributor:
    def __init__(self):
        self.id = None
        self.username = None
        self.ip = None

    def __repr__(self):
        if self.ip:
            return _repr(self, 'ip="%s"' % self.ip)
        return _repr(self, 'id=%s "%s"' % (self.id, self.username))


class Revision:
    def __init__(self):
        self.id = None
        self.parent_id = None
        self.timestamp = None
        self.contributor = None
        self.comment = None
        self.text = None
        self.text_length = None
        self.diff_length = None
        self.minor = False
        self.deleted_text = False
        self.deleted_contributor = False
        self.model = None
        self.format = None
        self.sha1 = None

    def __repr__(self):
        return _repr(self, 'id=%s m=%s' % (self.id, self.minor))


class Page:
    def __init__(self):
        self.title = None
        self.namespace = None
        self.id = None
        self.redirect = None
        self.restrictions = None
        self.revisions = []

    def __repr__(self):
        return _repr(self, 'title="%s"' % self.title)
